
(* :Title:  Quantum Algebra *)

(* :Context:  QuantumAlgebra`QuantumAlgebra` *)

(* :Author:  Cesar Guerra G.
             guerra_cesar@yahoo.com
             
             Pontificia Universidad Catolica del Peru 
             Av. Universitaria cdra. 18, San Miguel
             Lima-32 PERU. Telf. (511) 4602870 *)

(* :Summary:  Quantum Algebra is a package to perform Quantum Calculations
              using non commutative algebra. Dirac Notation is implemented. *)

(* :Mathematica Version: 4.1 and 4.2 *)

(* :Package Version:  1.0 *)

(* :Keywords: Quantum Mechanics, 
              Non Commutative Algebra, 
              Kets, Bras, Brakets, Dirac Notation *)

(* :Requirements:  Utilities`Notation` is needed to load QANotations.nb, 
                   a notebook which contains notations for some objects 
                   defined in this package.
                   If a notebook-based front end is not available notations
                   are not loaded and no requirements are needed. *)

(* :Warning:  1. QANotations.nb containing notations is evaluated from  
                 QuantumAlgebra.m calling SelectionEvaluateCreateCell[]
                 Since this command sends the evalution of the notebook
                 to the tail of all evaluations the kernel have to do, 
                 the command to load QuantumAlgebra must be evaluated 
                 alone in a single cell. 
              2. In this first version of Quantum Algebra warning and error
                 messages are not included, excepting usage messages. *)

(* :Discussion:  See QuantumAlgebra.nb demo. Several features and some 
                 examples are discussed. *)



BeginPackage["QuantumAlgebra`QuantumAlgebra`"]

Operator::usage =
"Operator is the head for an operator object. Some 
\nfull forms that can be interpreted are:
\n\t- Operator[A_], 
\n\t- Operator[A_,{x__},{}], 
\n\t- Operator[A_,{},{y__}],
\n\t- Operator[A_,{x__},{y__}] 
\nwhere A labels the operator and x are subscripts 
\nand y superscripts of the operator. 
\nIf y is an integer this is interpreted as a power.
\nOther combinations include 
\n\t- Operator[__][t__]
\n\t- Operator[A_,\"H\",__] 
\nwhere t are aditional variables and H is a label for 
\nthe associated hermitian operator of Operator[A]. 
\nOperator[1] and Operator[0] are interpreted as the 
\nidentity and cero operators respectively.
\nALIASES:
\n[ESC] op [ESC]  : operator.
\n[ESC] hop [ESC] : hermitian operator."

Ket::usage =
"Ket is the head for a Ket object. Some full forms that 
\ncan be interpreted are:
\n\t- Ket[A__]
\n\t- Ket[A__,{x__},{}], 
\n\t- Ket[A__,{},{y__}],
\n\t- Ket[A__,{x__},{y__}]
\nwhere A labels the Ket and x and y are subsripts and
\nsuperscripts respectively.
\nALIASES: 
\n[ESC] ket [ESC]   : ket
\n[ESC] ketd [ESC]  : ket with subscript
\n[ESC] ketu [ESC]  : ket with superscript
\n[ESC] ketdu [ESC] : ket with both scripts.
\nKet is also used as a tag for TaxBoxes that guide the 
\ninterpretation of the internal structure of ket notation."

Bra::usage =
"Bra is the head for a Bra object. Some full forms that 
\ncan be interpreted are:
\n\t- Bra[A__]
\n\t- Bra[A__,{x__},{}], 
\n\t- Bra[A__,{},{y__}],
\n\t- Bra[A__,{x__},{y__}]
\nwhere A labels the Ket and x and y are subsripts and
\nsuperscripts respectively.
\nALIASES: 
\n[ESC] bra [ESC]   : bra
\n[ESC] dbra [ESC]  : bra with subscript
\n[ESC] ubra [ESC]  : bra with superscript
\n[ESC] dubra [ESC] : bra with both scripts.
\nBra is also used as a tag for TaxBoxes that guide the 
\ninterpretation of the internal structure of bra notation."

Braket::usage =
"Braket is the head for a Braket object. Some full forms that 
\ncan be interpreted are:
\n\t- Braket[{A__},{B__}]
\n\t- Braket[{A__,{x__},{}},{B__}], 
\n\t- Braket[{A__,{},{y__}},{B__}],
\n\t- Braket[{A__,{x__},{y__}},{B__,{x__},{Y__}}]
\nwhere A and B label the Braket and x and y are subsripts 
\nand superscripts respectively. Of course more combinations
\nare posible.
\nALIASES: 
\n[ESC] braket [ESC]     : braket
\n[ESC] braketd [ESC]    : braket with subscript
\n[ESC] ubraket [ESC]    : braket with superscript
\n[ESC] dubraketdu [ESC] : braket with both scripts.
\nMore combinations are possible using u(up) and d(down).
\nBraket is also used as a tag for TaxBoxes that guide the 
\ninterpretation of the internal structure of braket notation."

Commutator::usage =
"Commutator[A,B] represents the commutator between operators A
\nand B. 
\nALIASES:
\n[ESC] com [ESC] : commutator."

Hermitian::usage = 
"Hermitian[expr] performs the hermitian operation on expr.
\nA expression with a dagger as a superscript is interpreted
\nas the hermiatian operation on expression"

QAPowerExpand::usage =
"QAPowerExpand[expr] expands all powers in expr as non 
\ncommutative products."

QAPowerContract::usage =
"QAPowerContract[expr] is intented to contract terms like: 
\nA^n ** A^m to A^(n+m) where A is an operator and n and m
\nsimple exponents."

QAProductExpand::usage =
"QAProductExpand[expr] expands all non commutative products of
\nconstant terms and quantum objects in expr as simple products
\nof constant terms and non commutative products of quantum
\nobjects."

QACommutatorExpand::usage =
"QACommutatorExpand[expr] expands all commutators in expr."

QACommutatorContract::usage =
"QACommutatorContract[expr] contracts all expanded commutators 
\nin expr. 
\nQACommutatorContract[expr,{A,B}] contracts all expanded form 
\nof Commutator[A,B] in expr. 
\nQACommutatorContract[expr,{{A,B},...}] performs several
\ncommutators contracts."

QAExpandAll::usage =
"QAExpandAll[expr] performs in expr the following commands:
\nQAPowerExpand, QAProductExpand, and QACommutatorExpand in that 
\norder"

QASeries::usage =
"QASeries[expr,{A,B,n}] expands the operator expr depending 
\nonly on A, in power series of A around the operator B up to
\nn order. The operator B must be of the form c Operator[1] or
\nOperator[0] (cero). 
\nQASeries[Exp[A] ** B ** Exp[-A],n] expands the first argument
\nas a series expansion in terms of A and B operators up to n 
\nterms.
\nQASeries does not return a series object, but a polinomial 
\non operaotor A."

QAOrdering::usage =
"QAOrdering[expr_, list_] transform expr which is formed by 
\nproduct of operators to an equivalent expression in which the
\norder of the products is changed. The new order will be as 
\ngiven in list from left to right."

BCHRelation::usage =
"BCHRelation[expr] use the Baker Campbell Hausdorff relation
\nto expand all operators of the form Exp[A+B] in expr where 
\nA and B are operators. BCHRelation automatically checks if 
\nA and B commute with their commutator."

QASet::usage =
"QASet[ expr1 = expr2] and QASet[expr1 := expr2] where expr1 and
\nexpr2 involve quantum objects like operators, bras, kets...,
\nfirst evaluate the Set and SetDelayed operations as given in 
\nits argument. After that, QASet try to set up definitions that
\ncan be deduced from the first relation. 
\nFor example if we evaluate QASet[ Commutator[A,B] = C ] then 
\nthe argument is evaluated in the usual way. But from this set 
\nit can be deduced that Commutator[B,A] = -C, so QASet also 
\ndoes this assignation.
\nQASet can handle the following definitions: 
\n\t- QASet[ Com[A,B] = expr ]
\n\t- QASet[ Com[A,Com[A,B]] = expr ]
\n\t- QASet[ Com[Com[A,B],Com[C,D]] = expr ]
\n\t- QASet[ Com[A,Com[A,B]] = 0, Com[B,Com[A,B]] = 0 ]
\n\t- QASet[ A ** Ket[n] = a Ket[n] ]
\nwhere A, B, C, D are operators and Com means Commutator. Also,
\nit can be used := instead of = if neccesary. For a full dicussion
\nsee QuantumAlgebra.nb Section 8. "

QAClear::usage =
"QAClear[ lhs =. ] removes all definitions made by QASet associated 
\nwith the definition of lhs. See QASet. "

SuperDagger::usage =
"Perform same operationa as Hermitian."

SuperStar::usage =
"Perform same operation as conjugate"

BraArgs::usage =
"BraArgs is a tag for TaxBoxes that guide the interpretation 
\nof the internal structure of bra notation."

KetArgs::usage =
"KetArgs is a tag for TaxBoxes that guide the interpretation 
\nof the internal structure of ket notation."

BraKetArgs::usage =
"BraKetArgs is a tag for TaxBoxes that guide the interpretation 
\nof the internal structure of braket notation."

UDScript::usage =
"UDScripts is a tag for TaxBoxes that guide the interpretation  
\nof the internal structure of bras, kets, and brakets notation."


Begin["`Private`"]

 protected = Unprotect[NonCommutativeMultiply, Exp, Times, Power]


 (***** DEFINING THE OBJECTS *****)

 SetAttributes[{OperatorQ, KetQ, BraQ, ConstantQ, QAProductExpand,
	QAPowerExpand, QACommutatorExpand, QAExpandAll}, Listable]

 ConstantQ[expr_] := FreeQ[expr,Operator] && FreeQ[expr,Ket] && FreeQ[expr,Bra]

 OperatorQ[t_] := !FreeQ[t, Operator]      

 KetQ[t_] := !FreeQ[t, Ket]
 
 BraQ[t_] := !FreeQ[t, Bra]

 
 (***** ALGEBRA FOR NON COMMUTATIVE MULTIPLY ****)

 0 ** P_ := 0
 
 P_ ** 0 := 0
 
 Operator[0] ** P_ := 0
 
 P_ ** Operator[0] := 0
 
 Operator/: Operator[0] + P_ := P
 
 Operator/: Operator[0] a_ := 0
 
 Operator[1] ** P_ := P
 
 P_ ** Operator[1] := P 
 
 P_ ** Q_Plus := Map[(P ** #)&, Q]

 Q_Plus ** P_ := Map[(# ** P)&, Q]

 P_ ** (c_ Q_) := c P ** Q /; ConstantQ[c]

 (c_ P_) ** Q_ := c P ** Q /; ConstantQ[c]

 
 (***** THE INNER PRODUCT *****)
 
 Bra[m__] ** Ket[n__] := Braket[{m},{n}]

 Braket[x__]**Braket[y__] := Braket[x] Braket[y]

 Braket/: Conjugate[Braket[{m__},{n__}]] := Braket[{n},{m}]


 (***** INTEGER SUPERSCRIPT INTERPRETATION AS POWER *****)

 Power[Operator[1], n_?IntegerQ ] := Operator[1]

 Power[Operator[0], n_?IntegerQ ] := 0

 Operator[P_,{},{n_?IntegerQ}] := Power[Operator[P],n]

 Power[Operator[P_],n_?IntegerQ][t__] := Power[Operator[P][t],n]

 Operator[P_,{x__},{n_?IntegerQ}] := Power[Operator[P,{x},{}],n]

 Power[Operator[P_,{x__},{}],n_?IntegerQ][t__] := Power[Operator[P,{x},{}][t],n]

 Power[expr_,0] := Operator[1] /; OperatorQ[expr] && expr =!= Operator[0] 

 Power[Operator[0],0] := Indeterminate

 Power/: P_^(-1) ** P_ := Operator[1] /; OperatorQ[P]

 Power/: P_ ** P_^(-1) := Operator[1] /; OperatorQ[P]


 (***** RULES FOR EXPONENTIAL OPERATORS *****)

 Power[E,a_. Operator[1]] := Exp[a] /; ConstantQ[a] 

 (* Power/: Power[E, P_] ** Power[E,d_.+ Q_] := Power[E,d] Power[E,P] ** 
					Power[E,Q] /; OperatorQ[P] && OperatorQ[Q] &&
							ConstantQ[c]&& ConstantQ[d] *)

 Power/: Power[E, P_] ** Power[E, Q_] := Operator[1] /; OperatorQ[P] && OperatorQ[Q] && P == -Q
 
 Power/: Power[E,c_. P_] ** Power[E,d_. P_] := Exp[(c+d) P] /; OperatorQ[P] 

 HoldPattern[Exp[a_. Operator[0]]] := Operator[1]

 (* Power/: Power[E,P_] ** Power[E,c_] := Exp[c] ** Exp[P] /; OperatorQ[P] && ConstantQ[c] *)

 SetAttributes[BCHRelation,HoldAllComplete]
 
 BCHRelation[expr_] := Module[ {sol}, ClearAttributes[Plus, Orderless] ;
                sol = expr /.  Power[E, c_. (P_ + Q_)]  :> 
		Module[{ com1 = c^2 Commutator[P,Q], com2 = c^2 Commutator[B,A]},        
		   If[Commutator[c P, com1] === Commutator[c Q, com1] === 0 ||
		      Commutator[c P, com2] === Commutator[c Q, com2] === 0 ||
		      Commutator[c P, com2] === Commutator[c Q, com1] === 0 ||
		      Commutator[c P, com1] === Commutator[c Q, com2] === 0 ||              
                      Commutator[c P, com1] === Commutator[com1, c Q] === 0 ||
		      Commutator[c P, com2] === Commutator[com2, c Q] === 0 ||
                      Commutator[c P, com1] === Commutator[com2, c Q] === 0 ||
		      Commutator[c P, com2] === Commutator[com1, c Q] === 0 ||
                      Commutator[com1, c P] === Commutator[c Q, com1] === 0 ||
		      Commutator[com1, c P] === Commutator[com1, c Q] === 0 ||
                      Commutator[com1, c P] === Commutator[c Q, com2] === 0 ||
		      Commutator[com2, c P] === Commutator[com1, c Q] === 0 ||	     
		      Commutator[com2, c P] === Commutator[c Q, com2] === 0 ||
		      Commutator[com2, c P] === Commutator[com2, c Q] === 0 ||
                      Commutator[com2, c P] === Commutator[c Q, com1] === 0 ||
		      Commutator[com1, c P] === Commutator[com2, c Q] === 0 ,
			QAProductExpand[ Exp[c P] ** Exp[c Q] ** Exp[-com1/2] ], 
			Exp[c(P + Q)]]]  /; OperatorQ[P] && OperatorQ[Q] && ConstantQ[c] ;
                SetAttributes[Plus,Orderless];
		sol]


 (***** RULES FOR HERMITIAN OPERATION *****)

 SuperDagger = Hermitian;

 SuperStar = Conjugate;

 Hermitian[Operator[0]] := Operator[0]

 Hermitian[Operator[1]] := Operator[1]

 Hermitian[Operator[P_,x___]] := Operator[P,"H",x]

 Hermitian[Operator[P_,"H",x___]] := Operator[P,x]

 Hermitian[c_] := Conjugate[c] /; ConstantQ[c]

 Hermitian[c_ P_] := Conjugate[c] Hermitian[P] /; ConstantQ[c] && !ConstantQ[P]

 Hermitian[ P_ + Q_] := Hermitian[P] + Hermitian[Q] /; !ConstantQ[P] && !ConstantQ[Q]

 Hermitian[ P_ ** Q_ ] := Hermitian[Q] ** Hermitian[P]

 Hermitian[Ket[a__]] := Bra[a]

 Hermitian[Bra[a__]] := Ket[a]

 Hermitian[Operator[P__][t__]] := Hermitian[Operator[P]][t]

 Hermitian[Power[P_,n_]] := Power[Hermitian[P],n] /; OperatorQ[P]

 Hermitian[Power[E,P_]] := Power[E,Hermitian[P]] /; OperatorQ[P]


 (***** RULES FOR COMMUTATION *****)

 Commutator[P_ , 0] := 0 

 Commutator[0, P_] := 0 

 Commutator[P_ ,Operator[0]] := 0

 Commutator[Operator[0],P_] := 0

 Commutator[P_ ,Operator[1]] := 0 /; OperatorQ[P]

 Commutator[Operator[1],P_] := 0 /; OperatorQ[P]

 Commutator[P_, P_] := 0 /; OperatorQ[P]

 Commutator[P_, n_ Q_] := n Commutator[P,Q] /; 
		!OperatorQ[n] && OperatorQ[P] && OperatorQ[Q] 
 
 Commutator[m_ P_, Q_] := m Commutator[P,Q] /; 
		!OperatorQ[m] && OperatorQ[P] && OperatorQ[Q]

 Commutator[P_, Q_+ R_] := Commutator[P,Q] + Commutator[P,R] /; 
		OperatorQ[P] && OperatorQ[Q] && OperatorQ[R]

 Commutator[P_+ Q_, R_] := Commutator[P,R] + Commutator[Q,R] /; 
		OperatorQ[P] && OperatorQ[Q] && OperatorQ[R]

 Commutator[P_, HoldPattern[Q_** R_]] := Commutator[P,Q] ** R + Q ** Commutator[P,R]/; 
		OperatorQ[P] && OperatorQ[Q] && OperatorQ[R]

 Commutator[HoldPattern[P_** Q_], R_] := Commutator[P,R] ** Q + P ** Commutator[Q,R]/; 
		OperatorQ[P] && OperatorQ[Q] && OperatorQ[R]

 Hermitian[HoldPattern[Commutator[P_, Q_]]] := Commutator[Hermitian[Q],Hermitian[P]]/; 
						OperatorQ[P] && OperatorQ[Q]

 Commutator[X:(Operator[__]|Operator[__][__]), expr_?OperatorQ]:= 0 /; checkOperator[X,expr]

 Commutator[expr_?OperatorQ,X:(Operator[__]|Operator[__][t__])] := 0 /; checkOperator[X,expr] 


 (***** RULES FOR QASet AND QAClear *****)

 SetAttributes[ QASet, {HoldAllComplete} ]
 SetAttributes[ QAClear, {HoldAllComplete} ]

 QASet[ Set[ ( Commutator[P_, Commutator[Q_,R_]] |
          Commutator[Commutator[R_,Q_], P_] ), S_ ] ] :=
                Module[ { }, 
		  Commutator[P,Commutator[Q,R]] = S;
                  Commutator[P,Commutator[R,Q]] = -S;
                  Commutator[Commutator[Q,R],P] = -S;
                  Commutator[Commutator[R,Q],P] = S; ]

 QASet[ Set[ Commutator[P_,Q_], R_] ] :=
		Module[ { x, y, auxP, auxQ},
		    Commutator[P,Q] = R;
                  Commutator[Q,P] = -R;
	           auxP = (P /. Verbatim[Pattern][x_, _] :> x); 
		    auxQ = (Q /. Verbatim[Pattern][x_, _] :> x);                  
                  If[P===auxP && Q===auxQ && Commutator[P,R]===Commutator[Q,R]===0,
                    Commutator[P^n_,Q^m_] /; IntegerQ[n] && IntegerQ[m] :=
			 Sum[Binomial[n,k] (m!/(m-k)!) (R^k) **
		              Q^(m-k)**P^(n-k),{k,1,n}]; 
		      Commutator[Q^n_,P^m_] /; IntegerQ[n] && IntegerQ[m] :=
			 Sum[Binomial[n,k] (m!/(m-k)!) ((-1)^k)(R^k) **
				P^(m-k)**Q^(n-k),{k,1,n}];
	 	    Commutator[P,x_] /; checkOperator[Q,x] := R**
				(D[x,Q]/.Times->NonCommutativeMultiply) //
                                                 MapAll[QAProductExpand,#]&;
		    Commutator[x_,P] /; checkOperator[Q,x] := -R**
				(D[x,Q]/.Times->NonCommutativeMultiply) //
                                                 MapAll[QAProductExpand,#]&;
		    Commutator[Q,y_] /; checkOperator[P,y] := -R**
				(D[y,P]/.Times->NonCommutativeMultiply) //
                                                 MapAll[QAProductExpand,#]&;
		    Commutator[y_,Q] /; checkOperator[P,y] := R**
				(D[y,P]/.Times->NonCommutativeMultiply) //
                                                 MapAll[QAProductExpand,#]&;
			]; ] /; R =!= 0
 		                   
 QASet[ Set[ Commutator[Commutator[P_,Q_],Commutator[R_,S_]], T_ ] ] :=
		Module[ { },
                  Commutator[Commutator[P,Q],Commutator[R,S]] = T;
                  Commutator[Commutator[P,Q],Commutator[S,R]] = -T;
                  Commutator[Commutator[Q,P],Commutator[R,S]] = -T;
                  Commutator[Commutator[Q,P],Commutator[S,R]] = T;
                  Commutator[Commutator[R,S],Commutator[P,Q]] = -T;
                  Commutator[Commutator[R,S],Commutator[Q,P]] = T;
                  Commutator[Commutator[S,R],Commutator[P,Q]] = T;
                  Commutator[Commutator[S,R],Commutator[Q,P]] = -T; ]

 QASet[ Set[ Commutator[P_,Q_], 0 ] ] :=
		Module[ { x, y },                  
		  Commutator[P,Q] = 0;
		  Commutator[Q,P] = 0;
		  Commutator[P,x_] /; checkOperator[Q,x] := 0;
	          Commutator[x_,P] /; checkOperator[Q,x] := 0;
		  Commutator[Q,x_] /; checkOperator[P,x] := 0;
		  Commutator[x_,Q] /; checkOperator[P,x] := 0; 
		  Commutator[x_,y_] /; checkOperator[P,x] && checkOperator[Q,y] := 0;
		  Commutator[x_,y_] /; checkOperator[Q,x] && checkOperator[P,y] := 0; ];

 QASet[ Set[ ( Commutator[P_,Commutator[P_,Q_]] |
	        Commutator[P_,Commutator[Q_,P_]] |
               Commutator[Commutator[P_,Q_],P_] |
	        Commutator[Commutator[Q_,P_],P_]  ), 0 ],
	 Set[ ( Commutator[Q_,Commutator[P_,Q_]] |
	        Commutator[Q_,Commutator[Q_,P_]] |
               Commutator[Commutator[P_,Q_],Q_] |
	        Commutator[Commutator[Q_,P_],Q_]  ), 0 ] ] :=
		 Module[ { x, y },	
		   Commutator[P,Commutator[P,Q]] = 0;
		   Commutator[P,Commutator[Q,P]] = 0;
		   Commutator[Commutator[P,Q],P] = 0;
		   Commutator[Commutator[Q,P],P] = 0;
		   Commutator[Q,Commutator[P,Q]] = 0;
		   Commutator[Q,Commutator[Q,P]] = 0;
		   Commutator[Commutator[P,Q],Q] = 0;
		   Commutator[Commutator[Q,P],Q] = 0;
		   Commutator[P^n_,Q^m_]  /; IntegerQ[n] && IntegerQ[m] := 
			Sum[Binomial[n,k] (m!/(m-k)!) (Commutator[P,Q]^k) **
				Q^(m-k)**P^(n-k),{k,1,n}];
		   Commutator[Q^n_,P^m_] /; IntegerQ[n] && IntegerQ[m] :=
			Sum[Binomial[n,k] (m!/(m-k)!) ((-1)^k)(Commutator[P,Q]^k) **
				P^(m-k)**Q^(n-k),{k,1,n}];
	  	   Commutator[P,x_] /; checkOperator[Q,x] := Commutator[P,Q]**
				(D[x,Q]/.Times->NonCommutativeMultiply) //
                                                     MapAll[QAProductExpand,#]&;
		   Commutator[x_,P] /; checkOperator[Q,x] := -Commutator[P,Q]**
				(D[x,Q]/.Times->NonCommutativeMultiply) //
                                                     MapAll[QAProductExpand,#]&;
		   Commutator[Q,y_] /; checkOperator[P,y] := -Commutator[P,Q]**
				(D[y,P]/.Times->NonCommutativeMultiply) //
                                                     MapAll[QAProductExpand,#]&;
		   Commutator[y_,Q] /; checkOperator[P,y] := Commutator[P,Q]**
				(D[y,P]/.Times->NonCommutativeMultiply) //
                                                     MapAll[ProductExpand,#]&; ]

 QASet[ Set[ P_?OperatorQ ** Ket[m__], c_. Ket[n__] ] ] :=
	Module[ {x,y = P}, 
		Unprotect[NonCommutativeMultiply];
	   	P ** Ket[m] = c Ket[n];
           	Bra[m] ** Hermitian[P] = Conjugate[c] Bra[n];
           If[ (Ket[m] /. Verbatim[Pattern][x_, _] :> x ) === Ket[n],
			x_ ** Ket[m] /; (OperatorQ[x] && checkOperator[P,x] && 
				Head[x] =!= NonCommutativeMultiply) := 
				     	QAProductExpand[( x Ket[n] /. y :> c )];
                   	Bra[m] ** x_ /; (OperatorQ[x] && checkOperator[Hermitian[P],x] &&
				Head[x] =!= NonCommutativeMultiply) := 
					QAProductExpand[( x Bra[n] /. 
						Hermitian[y] :> Conjugate[c])]; ]; 
		Protect[NonCommutativeMultiply]; ]

 QASet[ SetDelayed[ ( Commutator[P_, Commutator[Q_,R_]] |
          Commutator[Commutator[R_,Q_], P_] ), S_ ] ] :=
                Module[ { }, 
		  Commutator[P,Commutator[Q,R]] := S;
                  Commutator[P,Commutator[R,Q]] := -S;
                  Commutator[Commutator[Q,R],P] := -S;
                  Commutator[Commutator[R,Q],P] := S; ]

 QASet[ SetDelayed[ Commutator[P_,Q_], R_] ] :=
		Module[ { x, y, auxP, auxQ},
		  Commutator[P,Q] := R;
                  Commutator[Q,P] := -R;
	          auxP = (P /. Verbatim[Pattern][x_, _] :> x); 
		  auxQ = (Q /. Verbatim[Pattern][x_, _] :> x);                  
                  If[P===auxP && Q===auxQ && Commutator[P,R]===Commutator[Q,R]===0,
                    Commutator[P^n_,Q^m_] /; IntegerQ[n] && IntegerQ[m] :=
			 Sum[Binomial[n,k] (m!/(m-k)!) (R^k) **
		              Q^(m-k)**P^(n-k),{k,1,n}]; 
		    Commutator[Q^n_,P^m_] /; IntegerQ[n] && IntegerQ[m] :=
			 Sum[Binomial[n,k] (m!/(m-k)!) ((-1)^k)(R^k) **
				P^(m-k)**Q^(n-k),{k,1,n}];
	 	    Commutator[P,x_] /; checkOperator[Q,x] := R**
				(D[x,Q]/.Times->NonCommutativeMultiply) //
                                                  MapAll[QAProductExpand,#]&;
		    Commutator[x_,P] /; checkOperator[Q,x] := -R**
				(D[x,Q]/.Times->NonCommutativeMultiply) //
                                                  MapAll[QAProductExpand,#]&;
		    Commutator[Q,y_] /; checkOperator[P,y] := -R**
				(D[y,P]/.Times->NonCommutativeMultiply) //
                                                  MapAll[QAProductExpand,#]&;
		    Commutator[y_,Q] /; checkOperator[P,y] := R**
				(D[y,P]/.Times->NonCommutativeMultiply) //
                                                  MapAll[QAProductExpand,#]&;
			]; ] /; R =!= 0
 		                   
 QASet[ SetDelayed[ Commutator[Commutator[P_,Q_],Commutator[R_,S_]], T_ ] ] :=
		Module[ { },
                  Commutator[Commutator[P,Q],Commutator[R,S]] := T;
                  Commutator[Commutator[P,Q],Commutator[S,R]] := -T;
                  Commutator[Commutator[Q,P],Commutator[R,S]] := -T;
                  Commutator[Commutator[Q,P],Commutator[S,R]] := T;
                  Commutator[Commutator[R,S],Commutator[P,Q]] := -T;
                  Commutator[Commutator[R,S],Commutator[Q,P]] := T;
                  Commutator[Commutator[S,R],Commutator[P,Q]] := T;
                  Commutator[Commutator[S,R],Commutator[Q,P]] := -T; ]

 QASet[ SetDelayed[ Commutator[P_,Q_], 0 ] ] :=
		Module[ { x, y },                  
		  Commutator[P,Q] := 0;
		  Commutator[Q,P] := 0;
		  Commutator[P,x_] /; checkOperator[Q,x] := 0;
	          Commutator[x_,P] /; checkOperator[Q,x] := 0;
		  Commutator[Q,x_] /; checkOperator[P,x] := 0;
		  Commutator[x_,Q] /; checkOperator[P,x] := 0; 
		  Commutator[x_,y_] /; checkOperator[P,x] && checkOperator[Q,y] := 0;
		  Commutator[x_,y_] /; checkOperator[Q,x] && checkOperator[P,y] := 0; ];

 QASet[ SetDelayed[ ( Commutator[P_,Commutator[P_,Q_]] |
	        Commutator[P_,Commutator[Q_,P_]] |
               Commutator[Commutator[P_,Q_],P_] |
	        Commutator[Commutator[Q_,P_],P_]  ), 0 ],
	 Set[ ( Commutator[Q_,Commutator[P_,Q_]] |
	        Commutator[Q_,Commutator[Q_,P_]] |
               Commutator[Commutator[P_,Q_],Q_] |
	        Commutator[Commutator[Q_,P_],Q_]  ), 0 ] ] :=
		 Module[ { x, y },	
		   Commutator[P,Commutator[P,Q]] := 0;
		   Commutator[P,Commutator[Q,P]] := 0;
		   Commutator[Commutator[P,Q],P] := 0;
		   Commutator[Commutator[Q,P],P] := 0;
		   Commutator[Q,Commutator[P,Q]] := 0;
		   Commutator[Q,Commutator[Q,P]] := 0;
		   Commutator[Commutator[P,Q],Q] := 0;
		   Commutator[Commutator[Q,P],Q] := 0;
		   Commutator[P^n_,Q^m_]  /; IntegerQ[n] && IntegerQ[m] := 
			Sum[Binomial[n,k] (m!/(m-k)!) (Commutator[P,Q]^k) **
				Q^(m-k)**P^(n-k),{k,1,n}];
		   Commutator[Q^n_,P^m_] /; IntegerQ[n] && IntegerQ[m] :=
			Sum[Binomial[n,k] (m!/(m-k)!) ((-1)^k)(Commutator[P,Q]^k) **
				P^(m-k)**Q^(n-k),{k,1,n}];
	  	   Commutator[P,x_] /; checkOperator[Q,x] := Commutator[P,Q]**
				(D[x,Q]/.Times->NonCommutativeMultiply) //
                                                      MapAll[QAProductExpand,#]&;
		   Commutator[x_,P] /; checkOperator[Q,x] := -Commutator[P,Q]**
				(D[x,Q]/.Times->NonCommutativeMultiply) //
                                                      MapAll[QAProductExpand,#]&;
		   Commutator[Q,y_] /; checkOperator[P,y] := -Commutator[P,Q]**
				(D[y,P]/.Times->NonCommutativeMultiply) //
                                                      MapAll[QAProductExpand,#]&;
		   Commutator[y_,Q] /; checkOperator[P,y] := Commutator[P,Q]**
				(D[y,P]/.Times->NonCommutativeMultiply) //
                                                      MapAll[ProductExpand,#]&; ]

 QASet[ SetDelayed[ P_?OperatorQ ** Ket[m__], c_. Ket[n__] ] ] :=
	Module[ {x,y = P}, 
		Unprotect[NonCommutativeMultiply];
	   	P ** Ket[m] := c Ket[n];
           	Bra[m] ** Hermitian[P] := Conjugate[c] Bra[n];
           If[ (Ket[m] /. Verbatim[Pattern][x_, _] :> x ) === Ket[n],
			x_ ** Ket[m] /; (OperatorQ[x] && checkOperator[P,x] && 
				Head[x] =!= NonCommutativeMultiply) := 
				     	QAProductExpand[( x Ket[n] /. y :> c )];
                   	Bra[m] ** x_ /; (OperatorQ[x] && checkOperator[Hermitian[P],x] &&
				Head[x] =!= NonCommutativeMultiply) := 
					QAProductExpand[( x Bra[n] /. 
						Hermitian[y] :> Conjugate[c])]; ]; 
		Protect[NonCommutativeMultiply]; ]

 QAClear[ Unset[ ( Commutator[P_, Commutator[Q_,R_]] |
          Commutator[Commutator[R_,Q_], P_] )] ] :=
                Module[ { }, 
		    Commutator[P,Commutator[Q,R]] =.;
                  Commutator[P,Commutator[R,Q]] =.;
                  Commutator[Commutator[Q,R],P] =.;
                  Commutator[Commutator[R,Q],P] =.; ]

 QAClear[ Unset[ Commutator[P_,Q_] ] ] :=
		Module[ { x, y, auxP, auxQ, T = Commutator[P,Q],
				R = Commutator[P,Commutator[P,Q]], 
				S = Commutator[Q,Commutator[P,Q]]},
		    Commutator[P,Q] =.;
                  Commutator[Q,P] =.;
	           auxP = (P /. Verbatim[Pattern][x_, _] :> x); 
		    auxQ = (Q /. Verbatim[Pattern][x_, _] :> x);                  
                  If[P===auxP && Q===auxQ && R===S===0 && T=!=0,
                    	Commutator[P^n_,Q^m_] /; IntegerQ[n] && IntegerQ[m] =.;
		      	Commutator[Q^n_,P^m_] /; IntegerQ[n] && IntegerQ[m] =.;
	 	      	Commutator[P,x_] /; checkOperator[Q,x] =.;
		      	Commutator[x_,P] /; checkOperator[Q,x] =.;
		      	Commutator[Q,y_] /; checkOperator[P,y] =.;
		      	Commutator[y_,Q] /; checkOperator[P,y] =.; ];
		    If[T===0,
			Commutator[P,x_] /; checkOperator[Q,x] =.;
	          	Commutator[x_,P] /; checkOperator[Q,x] =.;
		  	Commutator[Q,x_] /; checkOperator[P,x] =.;
		  	Commutator[x_,Q] /; checkOperator[P,x] =.;
		  	Commutator[x_,y_] /; checkOperator[P,x] && checkOperator[Q,y] =.;
			Commutator[x_,y_] /; checkOperator[Q,x] && checkOperator[P,y] =. ]; ]

 QAClear[ Unset[ Commutator[Commutator[P_,Q_],Commutator[R_,S_]] ] ] :=
		Module[ { },
                  Commutator[Commutator[P,Q],Commutator[R,S]] =.;
                  Commutator[Commutator[P,Q],Commutator[S,R]] =.;
                  Commutator[Commutator[Q,P],Commutator[R,S]] =.;
                  Commutator[Commutator[Q,P],Commutator[S,R]] =.;
                  Commutator[Commutator[R,S],Commutator[P,Q]] =.;
                  Commutator[Commutator[R,S],Commutator[Q,P]] =.;
                  Commutator[Commutator[S,R],Commutator[P,Q]] =.;
                  Commutator[Commutator[S,R],Commutator[Q,P]] =.; ]

 QAClear[ Unset[ ( Commutator[P_,Commutator[P_,Q_]] |
	        Commutator[P_,Commutator[Q_,P_]] |
               Commutator[Commutator[P_,Q_],P_] |
	        Commutator[Commutator[Q_,P_],P_]  ) ],
	   Unset[ ( Commutator[Q_,Commutator[P_,Q_]] |
	        Commutator[Q_,Commutator[Q_,P_]] |
               Commutator[Commutator[P_,Q_],Q_] |
	        Commutator[Commutator[Q_,P_],Q_]  )  ] ] :=
		 Module[ { x, y },	
		   Commutator[P,Commutator[P,Q]] =.;
		   Commutator[P,Commutator[Q,P]] =.;
		   Commutator[Commutator[P,Q],P] =.;
		   Commutator[Commutator[Q,P],P] =.;
		   Commutator[Q,Commutator[P,Q]] =.;
		   Commutator[Q,Commutator[Q,P]] =.;
		   Commutator[Commutator[P,Q],Q] =.;
		   Commutator[Commutator[Q,P],Q] =.;
		   Commutator[P^n_,Q^m_] /; IntegerQ[n] && IntegerQ[m] =.;
		   Commutator[Q^n_,P^m_] /; IntegerQ[n] && IntegerQ[m] =.;
	  	   Commutator[P,x_] /; checkOperator[Q,x] =.;
		   Commutator[x_,P] /; checkOperator[Q,x] =.;
		   Commutator[Q,y_] /; checkOperator[P,y] =.;
		   Commutator[y_,Q] /; checkOperator[P,y] =.; ]

 QAClear[ Unset[ P_?OperatorQ ** Ket[m__] ] ] :=
	Module[ {x,y = P}, 
		Unprotect[NonCommutativeMultiply];
	   	P ** Ket[m] =.;
           	Bra[m] ** Hermitian[P] =.;
		Off[Unset::norep];
		x_ ** Ket[m] /; (OperatorQ[x] && checkOperator[P,x] && 
				Head[x] =!= NonCommutativeMultiply) =.;
                Bra[m] ** x_ /; (OperatorQ[x] && checkOperator[Hermitian[P],x] &&
				Head[x] =!= NonCommutativeMultiply) =.;
		On[Unset::norep]; 
		Protect[NonCommutativeMultiply]; ] 


 (***** RULES FOR EXPANDING SERIES *****)

 QASeries[expr_, {x_,a_, n_}] := Module[{k,t,f},
		t = (((expr /. x->a) /. Operator[1]->1) Operator[1] +
		       Sum[(( (D[expr,{x,k}] /. x->a) /. 
			    Operator[1]->1) (x-a)^k) / f[k], {k,1,n}]) /. Operator[0] -> 0;
		t = t /. f[v_] :> HoldForm[v!]]

 QASeries[c_. Exp[P_] ** R_ ** Exp[Q_], n_] :=
		Module[{f, k, s},
		    s = c Sum[Nest[Commutator[P,#]&, R, k]/f[k], {k, 0, n-1}];
		    s = s /. f[v_] :> HoldForm[v!]] /; 
			OperatorQ[P] && OperatorQ[R] && P === -Q && Commutator[P,R] =!= 0



 (***** RULES FOR SIMPLIFYING QUANTUM OBJECTS *****)

 QAOrdering[ expr_, list_List ] := expr //. CommutationOrderRules[list]

 QAPowerExpand[expr_] := expr //. rule1

 QAPowerContract[expr_] := expr //. rule2

 QACommutatorExpand[expr_?OperatorQ] := expr //. rule3 
 
 QACommutatorExpand[expr_] := expr

 QAProductExpand[expr_?(!ConstantQ[#]&)] := expr //. rule4
 
 QAProductExpand[expr_?ConstantQ] := expr //. NonCommutativeMultiply -> Times
 
 QAProductExpand[expr_] := expr

 QAExpandAll[expr_?OperatorQ] := expr // QAPowerExpand // QAProductExpand //
					QACommutatorExpand //Expand

 QAExpandAll[expr_] := expr

 QACommutatorContract[expr_, t : {{_, _} ...}] := 
	Fold[ReplaceRepeated, expr,                        			
		Map[First[#] ** Last[#] :>
                           Commutator[First[#], Last[#]] + Last[#] ** First[#] &, t]]

 QACommutatorContract[expr_,{P_,Q_}] := expr //. P**Q :> Commutator[P,Q] + Q**P

 
 QACommutatorContract[expr_] := expr //. P_**Q_ - Q_**P_ :> Commutator[P,Q]

 
 (***** Auxiliar Definitions and rules *****)

 CommutationOrderRules[l_List] := 
	Module[{x = {First[l]}, y = Rest[l], f, sol = {}},
	   f = Rule[#2**#1, #1**#2 + Commutator[#2, #1]] &;
	   While[Length[y] >= 1, 
		   AppendTo[sol, Outer[f, x, y]];
	    	   x = {First[y]};
		   y = Rest[y]];
		   Flatten[sol]]

 expand[ NonCommutativeMultiply[P__]] :=
	Module[{temp1, temp2, l1,  temp3 = {P}},
		temp1 = Select[temp3,!ConstantQ[#]&];	
		temp2 = Select[temp3, ConstantQ[#]&];
		l1 = Length[temp1];
	        Which[l1 > 1, Times[ Apply[NonCommutativeMultiply,temp1],Apply[Times,temp2]],
	      	      l1===1, First[temp1] Apply[Times,temp2],
	      	      l1===0, Apply[Times,temp2]]]

 rule1 = {(P_Plus)^n_?IntegerQ /; OperatorQ[P] && n > 0 :> Plus @@ Flatten[
                      Outer[NonCommutativeMultiply,Sequence @@ Table[List @@ P,{n}]]],
          Power[P_?OperatorQ, n_?IntegerQ] /; n > 0 :>
		     Apply[NonCommutativeMultiply, Table[P,{n}]],
          Power[P_?OperatorQ, n_?IntegerQ] /; n < -1 :>
		     Apply[NonCommutativeMultiply, Table[P^(-1),{-n}]]}

 rule2 = P_^n_. ** P_^m_. :> P^(n+m) /; OperatorQ[P]

 rule3 = HoldPattern[Commutator[P_, Q_]] :> P ** Q - Q ** P

 rule4 = NonCommutativeMultiply[P__] :> expand[NonCommutativeMultiply[P]]

 checkOperator[Operator[P__],expr_] :=
    Complement[Cases[expr, (Operator[__]|Operator[__][__]), {0, Infinity}], 
	      {Operator[0],Operator[1]} ] ===
	      Union[Cases[expr, Operator[P], {0, Infinity}]] && expr =!= Operator[P]

 checkOperator[Operator[P__][t__],expr_] :=
    Complement[Cases[expr, (Operator[__]|Operator[__][__]), {0, Infinity}],
	      {Operator[0],Operator[1]} ] ===
            Union[Cases[expr, Operator[P][t], {0, Infinity}]] && expr =!= Operator[P][t]

 (***** Load Notations *****)

 If[$Notebooks,
        NotebookOpen["QAPalette.nb"];
	nb1 = EvaluationNotebook[];
	nb2 = NotebookOpen["QANotations.nb", Visible->False];
        nb3 = NotebookCreate[ 
			WindowMovable->False,
			WindowFloating->True,
			WindowSize->{210, 60}, 
			WindowMargins->{{113, Automatic}, {161, Automatic}},
			WindowFrame->"Palette", 
			WindowElements->{}, 
			WindowFrameElements->"CloseBox",
			WindowTitle->"QAMessage", 
			Background-> RGBColor[0.816419, 0.734386, 0.261723],
			ShowCellBracket->False];
	SelectionMove[nb2, All, Notebook];
	SelectionEvaluateCreateCell[nb2];
  ] 

 
 Protect[ Release[protected] ]

 End[]

 EndPackage[]










